unit signalanalu;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  ExtCtrls, Menus, ActnList, ComCtrls, fftw_s;

type

  Tanzeigetyp=(atspektrum,atpolar);

  { TSignalanalyseF }

  TSignalanalyseF = class(TForm)
    AAFK: TAction;
    MenuItem5: TMenuItem;
    MenuItem6: TMenuItem;
    MenuItem7: TMenuItem;
    MenuItem8: TMenuItem;
    MenuItem9: TMenuItem;
    PolarA: TAction;
    SpektrumA: TAction;
    CloseA: TAction;
    RepaintTimer: TIdleTimer;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    SaveA: TAction;
    OpenA: TAction;
    ActionList1: TActionList;
    MainMenu1: TMainMenu;
    PaintBox1: TPaintBox;
    procedure AAFKExecute(Sender: TObject);
    procedure CloseAExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MenuItem1Click(Sender: TObject);
    procedure OpenAExecute(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBox1MouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure PaintBox1Paint(Sender: TObject);
    procedure PolarAExecute(Sender: TObject);
    procedure RepaintTimerTimer(Sender: TObject);
    procedure SaveAExecute(Sender: TObject);
    procedure SpektrumAExecute(Sender: TObject);
  private
    { private declarations }
    links,rechts:integer;
    fft_i,fft_o:Pcomplex_single;
    fft_l:integer;
    fft_plan:fftw_plan_single;
    fft_binwidth:double; //Breite eines FFT Bins
    power:ARRAY OF ARRAY of single;
    minlog,maxlog:double;
    anzeigetyp:Tanzeigetyp;
    amaus:tPoint;
    auswahl1,auswahl2:tPoint;
    FUNCTION xtosample(x:integer):integer;
    FUNCTION ytofreq(y:integer):double;
    PROCEDURE loadfromfile(fn:string);
    PROCEDURE savetofile(fn:string);
    PROCEDURE dofft(fr,tm:integer);
    PROCEDURE paintspektrum(canv:TCanvas);
    PROCEDURE paintpolar(canv:TCanvas);
    PROCEDURE PaintZSkala(canv:TCanvas);
    PROCEDURE downsample(s1,s2:integer; f1,f2:double);
  public
    { public declarations }
    samplerate:double;
    midfrequency:double;
    starttime:double;
    samples:ARRAY OF complex_single;
  end; 

var
  SignalanalyseF: TSignalanalyseF;

implementation

uses
  BGRABitmap, BGRABitmapTypes,vermischtesU;

PROCEDURE TSignalanalyseF.dofft(fr,tm:integer);
VAR n,m,sp,asp:integer;
    l,h,w:integer;
    x:integer;
    lv:double; //ln(wert)
    f:double;
BEGIN
  fr:=finde2potenz(fr);
  w:=fr;
  IF fft_l<>fr THEN //Wenn die FFT-Laengen anders sind, anderen Plan machen
  BEGIN
    IF fft_plan<>NIL THEN fftw_destroy_plan(fft_plan);
    IF fft_i<>NIL THEN fftw_freemem(fft_i);
    IF fft_o<>NIL THEN fftw_freemem(fft_o);
    fftw_getmem(fft_i,fr*sizeof(complex_single));
    fftw_getmem(fft_o,fr*sizeof(complex_single));
    fft_plan:=fftw_plan_dft_1d(fr,fft_i,fft_o,fftw_forward,[fftw_estimate]);
    fft_l:=fr;
  end;
  fft_binwidth:=samplerate/fft_l;
  minlog:=maxint;
  maxlog:=-maxint;
  SetLength(power,tm,fr);
  minlog:=maxint;
  maxlog:=-maxint;
  asp:=-1;
  FOR n:=0 TO tm-1 DO
  BEGIN
    sp:=round((n/tm)*(rechts-links)+links);
    IF sp<>asp THEN //Falls zu stark reingezoomt, nicht spektrum mehrmals berechnen.
    BEGIN
      l:=low(samples);
      h:=High(samples);
      FOR m:=-w div 2 TO w div 2-1 DO
      BEGIN
        IF (sp+m<l) OR (sp+m>h) THEN
        BEGIN
          fft_i[m+w div 2].re:=0;
          fft_i[m+w div 2].im:=0;
        end ELSE
        BEGIN
          f:=1; //0.54+cos(m/w*pi)*0.46;  //Hier Fensterung einfügen falls gewünscht
          fft_i[m+w div 2].re:=samples[sp+m].re*f;
          fft_i[m+w div 2].im:=samples[sp+m].im*f;
        end;
      end;
      fftw_execute(fft_plan);
      asp:=sp;
    end;
    FOR m:=0 TO fr-1 DO
    BEGIN
      lv:=abs(fft_o[m].re)+abs(fft_o[m].im);
      IF lv<10/sqr(fft_l) THEN lv:=10/sqr(fft_l);
      power[n,m]:=lv;
      lv:=ln(lv);
      IF lv>maxlog THEN maxlog:=lv;
      IF lv<minlog THEN minlog:=lv;
    end;
  end;
end;




PROCEDURE TSignalanalyseF.paintspektrum(canv:TCanvas);
var
  X, Y: Integer;
  P: PBGRAPixel;
  bgra: TBGRABitmap;
  w,h:integer;
  f:integer;
  v:double;
  fr:integer;
  n,m:integer;
  fs:double;
  fu,fo:double;
  s:string;
BEGIN
  w:=Canv.Width;
  h:=Canv.Height;
//  w:=h;//128;
 //<  h:=256;
  IF rechts<=links THEN exit;
  dofft(h*3 div 2,w);
  bgra := TBGRABitmap.Create(w,h);
  for Y := 0 to h - 1 do
  begin
    P := PBGRAPixel(bgra.ScanLine[Y]);
    for X := 0 to w - 1 do
    begin
      fr:=(h div 2-y);
      v:=power[x,(fr+2*fft_l) MOD fft_l];
      IF v<exp(minlog) THEN v:=exp(minlog);
      v:=ln(v);
      f:=round((v-minlog)/(maxlog-minlog)*256);
      IF f<0 THEN f:=0;
      IF f>255 THEN f:=255;
        p^.red:=f;
        p^.green:=f;
        p^.Blue:=f;
      p^.Alpha:=255;
      // It is a shortcut for :
      // P^.Red := Pixels[X, Y];
      // P^.Green := Pixels[X, Y];
      // P^.Blue := Pixels[X, Y];
      // P^.Alpha := 255;
      Inc(P);
    end;
  end;
  bgra.InvalidateBitmap; // Changed by direct access
  bgra.Draw(Canv, 0, 0, false);
  bgra.Free;
  //Frequenzskala
  fu:=midfrequency-fft_binwidth*h/2;
  fo:=midfrequency+fft_binwidth*h/2;
  y:=Canv.TextHeight('Tg') div 2;
  fs:=(fo-fu)/h*y*5; //Frequenzschritt
  fs:=finderundezahl(fs); //Runden
  FOR n:=round(fu/fs) TO round(fo/fs) DO
  BEGIN
    f:=h-round((n*fs-fu)/(fo-fu)*h);
    s:=FloatToStr(n*fs);
    x:=50-Canv.TextWidth(s);
    Canv.Font.Color:=clWhite;
    Canv.TextOut(x-1,f-y-1,s);
    Canv.Font.Color:=clBlack;
    Canv.TextOut(x+1,f-y+1,s);
    Canv.Font.Color:=clRed;
    Canv.TextOut(x,f-y,s);
    Canv.Pen.Color:=clWhite;
    FOR m:=0 TO 5 DO
      Canv.Line(50+m*10,f,54+m*10,f);
    Canv.Pen.Color:=clBlack;
    FOR m:=0 TO 5 DO
      Canv.Line(55+m*10,f,59+m*10,f);
  end;
end;

PROCEDURE TSignalanalyseF.PaintZSkala(canv:TCanvas);
VAR fu,fo,fs:double;
    x,y,h,w,n,f:integer;
    s:string;
BEGIN
  h:=Canv.Height;
  w:=Canv.Width;
//Zeitskala
  y:=Canv.TextHeight('Tg') div 2;
  fu:=starttime+links*1/samplerate;
  fo:=starttime+rechts*1/samplerate;
  x:=Canv.TextWidth('00:00:00') div 2;
  fs:=(fo-fu)/w*x*5;
  fs:=finderundezeit(fs);
  FOR n:=round(fu/fs) TO round(fo/fs) DO
  BEGIN
    f:=round((n*fs-fu)/(fo-fu)*w);
    s:=zeitinstring((n*fs));
    x:=f-Canv.TextWidth(s);
    Canv.Font.Color:=clWhite;
    Canv.TextOut(x-1,h-y*3-1,s);
    Canv.Font.Color:=clBlack;
    Canv.TextOut(x+1,h-y*3+1,s);
    Canv.Font.Color:=clRed;
    Canv.TextOut(x,h-y*3,s);
    Canv.Pen.Color:=clWhite;
  end;
end;

FUNCTION extractfromstring(s,name:string):string;
VAR p:integer;
BEGIN
  p:=pos(name,s);
  result:=trim(copy(s,p+1+length(name),maxint));
end;

procedure TSignalanalyseF.OpenAExecute(Sender: TObject);
VAR fm:TOpenDialog;
begin
  fm:=TOpenDialog.Create(Application);
  TRY
  fm.FileName:='/tmp/temp/temp.dat';
    fm.Filter:='*.dat';
    IF fm.Execute THEN
    loadfromfile(fm.FileName);
  finally
    fm.free;
  end;
end;

procedure TSignalanalyseF.PaintBox1MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  IF (Button=mbRight) AND (anzeigetyp=atspektrum) THEN
  BEGIN
    IF auswahl1.x<0 THEN
    BEGIN
      auswahl1.x:=x; auswahl1.y:=y;
      auswahl2.x:=-1;
    end ELSE
    BEGIN
      auswahl2.x:=x; auswahl2.y:=y;
      downsample(xtosample(auswahl1.x),xtosample(auswahl2.x),ytofreq(auswahl1.y),ytofreq(auswahl2.y));
      auswahl1.x:=-1;
    end;
  end;
end;

procedure TSignalanalyseF.PaintBox1MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
  PROCEDURE zeichne;
  BEGIN
    PaintBox1.Canvas.Line(0,amaus.y,PaintBox1.ClientWidth,amaus.y);
    PaintBox1.Canvas.Line(amaus.x,0,amaus.x,PaintBox1.ClientHeight);
    IF auswahl1.x>=0 THEN
    PaintBox1.Canvas.Rectangle(auswahl1.x,auswahl1.y,amaus.x,amaus.y);
  end;

begin
  PaintBox1.Canvas.Pen.Mode:=pmXor;
  PaintBox1.Canvas.Pen.color:=clWhite;
  IF amaus.x>=0 THEN
  BEGIN
    zeichne;
  end;
  amaus.x:=x;
  amaus.y:=y;
  zeichne;
  PaintBox1.Canvas.Pen.Mode:=pmCopy;
end;

procedure TSignalanalyseF.PaintBox1MouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
VAR p,l,r:integer;
begin
  p:=round(MousePos.x/PaintBox1.ClientWidth*(rechts-links)+links);
  l:=links;
  r:=rechts;
  WHILE WheelDelta<>0 DO
  BEGIN
    IF WheelDelta>0 THEN
    BEGIN
      l:=round((l-p)*0.999+p+1);
      r:=round((r-p)*0.999+p-1);
      dec(WheelDelta);
    end;
    IF WheelDelta<0 THEN
    BEGIN
      l:=round((l-p)*1.001+p-1);
      r:=round((r-p)*1.001+p+1);
      inc(WheelDelta);
    end;
    If r-l<100 THEN
    BEGIN
      r:=r+50;
      l:=l-50;
    END;
    IF l<Low(Samples) THEN l:=Low(Samples);
    IF r>High(Samples) THEN r:=High(Samples);
  end;
  links:=l; rechts:=r;
  //PaintBox1.Refresh;
  RepaintTimer.Enabled:=true;
end;


PROCEDURE TSignalanalyseF.paintpolar(canv:TCanvas);
VAR mag,phase:ARRAY OF single;
    n:integer;
    x,y,ax,ay:integer;
    ymax,ymin:integer;
    maxmag:double;
BEGIN
  SetLength(mag,rechts-links+1);
  SetLength(phase,rechts-links+1);
  maxmag:=0;
  FOR n:=0 TO rechts-links DO
  BEGIN
    mag[n]:=sqrt(sqr(samples[n+links].re)+sqr(samples[n+links].im));
    phase[n]:=winkel(samples[n+links].re,samples[n+links].im);
    IF mag[n]>maxmag THEN maxmag:=mag[n];
  end;
  ax:=-1;
  ay:=0;
  ymax:=0;
  ymin:=0;
  canv.Pen.Color:=clblack;
  FOR n:=0 TO rechts-links DO
  BEGIN
    x:=round(n/(rechts-links+1)*canv.width);
    y:=canv.height-round(mag[n]/maxmag*canv.height);
    IF x<>ax THEN
    BEGIN
      canv.line(ax,ay,x,y);
      IF ymax>ymin THEN canv.line(ax,ymin,ax,ymax);
      ax:=x;
      ay:=y;
      ymax:=y;
      ymin:=y;
    end ELSE
    BEGIN
      IF y>ymax THEN ymax:=y;
      IF y<ymin THEN ymin:=y;
    end;
  end;
  ax:=-1;
  ay:=0;
  ymax:=0;
  ymin:=0;
  canv.Pen.Color:=clblue;
  FOR n:=0 TO rechts-links DO
  BEGIN
    x:=round(n/(rechts-links+1)*canv.width);
    y:=canv.height-round((phase[n]/pi/2+0.5)*canv.height);
    IF x<>ax THEN
    BEGIN
      canv.line(ax,ay,x,y);
      IF ymax>ymin THEN canv.line(ax,ymin,ax,ymax);
      ax:=x;
      ay:=y;
      ymax:=y;
      ymin:=y;
    end ELSE
    BEGIN
      IF y>ymax THEN ymax:=y;
      IF y<ymin THEN ymin:=y;
    end;
  end;
end;

procedure TSignalanalyseF.PaintBox1Paint(Sender: TObject);
begin
  IF anzeigetyp=atspektrum THEN paintspektrum(PaintBox1.Canvas) ELSE
  IF anzeigeTyp=atpolar    THEN paintpolar(PaintBox1.Canvas);
  PaintZSkala(PaintBox1.Canvas);
  amaus.x:=-1; amaus.y:=-1;
  auswahl1.x:=-1;
end;

procedure TSignalanalyseF.PolarAExecute(Sender: TObject);
begin

end;

procedure TSignalanalyseF.RepaintTimerTimer(Sender: TObject);
begin
  PaintBox1.Refresh;
  (Sender as TIdleTimer).enabled:=false;
end;

procedure TSignalanalyseF.MenuItem1Click(Sender: TObject);
begin

end;

procedure TSignalanalyseF.CloseAExecute(Sender: TObject);
begin
  close;
end;

procedure TSignalanalyseF.AAFKExecute(Sender: TObject);
begin

end;

procedure TSignalanalyseF.FormCreate(Sender: TObject);
begin
  fft_l:=-1;
  fft_i:=NIL;
  fft_o:=NIL;
  fft_plan:=NIL;
  anzeigetyp:=atspektrum;
  amaus.x:=-1;
  amaus.y:=-1;
   auswahl1.x:=-1;
end;

procedure TSignalanalyseF.SaveAExecute(Sender: TObject);
VAR fm:TSaveDialog;
begin
  fm:=TSaveDialog.Create(Application);
  TRY
    fm.Filter:='*.dat';
    IF fm.Execute THEN savetofile(fm.FileName);
  finally
    fm.free;
  end;
end;

procedure TSignalanalyseF.SpektrumAExecute(Sender: TObject);
begin
  IF Sender=SpektrumA THEN anzeigetyp:=atspektrum ELSE
  IF Sender=PolarA    THEN anzeigetyp:=atPolar;
  PaintBox1.Refresh;
end;

PROCEDURE TSignalanalyseF.loadfromfile(fn:string);
VAR f:textfile;
    s:string;
    p:integer;
    channels:integer;
    n:integer;
    t:double;
BEGIN
  links:=0;
  rechts:=-1;
  SetLength(samples,0);
  assignfile(f,fn);
  reset(f);
  readln(f,s);
  samplerate:=strtofloat(extractfromstring(s,'Sample Rate'));
  readln(f,s);
  channels:=StrToInt(extractfromstring(s,'Channels'));
  n:=0;
  REPEAT
     IF n>High(samples) THEN SetLength(Samples,n*2+1);
     IF channels=1 THEN
     BEGIN
       readln(f,t,Samples[n].re);
       samples[n].im:=0;
     end ELSE
     BEGIN
       readln(f,t,Samples[n].re,samples[n].im);
     end;
     IF n=0 THEN starttime:=t;
     inc(n);
  until eof(f);
  SetLength(samples,n);
  closefile(f);
  links:=Low(samples);
  rechts:=High(samples);
  Caption:=fn+' '+inttostr(n);
  PaintBox1.Refresh;
end;

PROCEDURE TSignalanalyseF.savetofile(fn:string);
VAR f:textfile;
    n:integer;
BEGIN
  AssignFile(f,fn);
  Rewrite(f);
  writeln(f,'; Sample Rate ',samplerate);
  writeln(f,'; Channels 2');
  FOR n:=Low(samples) TO High(samples) DO
  BEGIN
    writeln(f,starttime+n/samplerate:10,' ',samples[n].re:8,' ',samples[n].im:8);
  end;
  CloseFile(f);
end;

FUNCTION TSignalanalyseF.xtosample(x:integer):integer;
BEGIN
  result:=round((x/PaintBox1.ClientWidth)*(rechts-links)+links);
END;

FUNCTION TSignalanalyseF.ytofreq(y:integer):double;
BEGIN
  result:=(PaintBox1.ClientHeight/2-y)*fft_binwidth+midfrequency;
end;


PROCEDURE TSignalanalyseF.downsample(s1,s2:integer; f1,f2:double);
VAR form:TSignalanalyseF;
    n,m,a,b,c:integer;
    mf,bw,w:double;
    sumr,sumi:double;
    phi:double;
BEGIN
  If f2<f1 THEN
  BEGIN
    downsample(s1,s2,f2,f1);
    exit;
  end;
  IF s2<s1 THEN
  BEGIN
    downsample(s2,s1,f1,f2);
    exit;
  end;
  form:=TSignalanalyseF.Create(Application);
  mf:=(f1+f2)/2;
  bw:=(f2-f1);
  form.midfrequency:=mf;
  form.samplerate:=bw*1.5;
  n:=round((s2-s1)/samplerate*form.samplerate);
  Caption:=inttostr(n);
  SetLength(form.samples,n);
  Caption:=inttostr(low(form.samples))+' '+inttostr(high(form.samples));
  FOR n:=Low(form.samples) TO High(form.samples) DO
  BEGIN    //Heruntermischen komplex!
    c:=round(n/form.samplerate*samplerate+s1);
    a:=c-10;
    b:=c+10;
    IF a<Low(samples) THEN a:=Low(samples);
    IF b>High(samples) THEN b:=High(samples);
    sumr:=0; sumi:=0;
    FOR m:=a TO b DO
    BEGIN
      w:=si((m-c)/samplerate*bw*pi);
      phi:=m/samplerate*2*pi*(midfrequency-mf);
      sumr:=sumr+ samples[m].re*cos(phi)-samples[m].im*sin(phi);
      sumi:=sumi+ samples[m].re*sin(phi)+samples[m].im*cos(phi);
    end;
    form.samples[n].re:=sumr;
    form.samples[n].im:=sumi;
  end;
  form.links:=Low(form.samples);
  form.rechts:=High(form.samples);
  form.Show;
end;

initialization
  {$I signalanalu.lrs}

end.

